# -*- coding: utf-8 -*-
#
# satextrato/base.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import textwrap
import warnings

from unidecode import unidecode

import escpos.barcode
import escpos.feature

from satcomum import br
from satcomum import constantes
from satcomum import util

from .config import conf


def _bordas(esquerda, direita, largura=48, espacamento_minimo=4):
        espacamento = largura - (len(esquerda) + len(direita))
        if espacamento < espacamento_minimo:
            espacamento = espacamento_minimo
            comprimento_maximo = int(largura / 2) - espacamento_minimo
            esquerda = esquerda[:comprimento_maximo]
            direita = direita[:comprimento_maximo]
        return '%s%s%s' % (esquerda, ' ' * espacamento, direita)


class ExtratoCFe(object):
    """
    Classe base para os extratos CF-e, fornecendo uma implementação padrão
    para o cabeçalho dos CF-e de venda e para os CF-e de cancelamento. Também
    fornece uma infraestrutura que simplifica a interface com impressoras
    ESC/POS, acrescentando fluidez à API de impressão.
    """

    def __init__(self, xmlstr, impressora):
        """

        :param xmlstr: String (unicode) contendo o XML do CF-e-SAT.
        :param impressora: Um objeto :class:`escpos.impl.epson.GenericESCPOS`
            ou especialização.
        """
        super(ExtratoCFe, self).__init__()
        self._xmlstr = xmlstr
        self.xml = util.XMLFacadeFromString(self._xmlstr)
        self.impressora = impressora

        self._flag_negrito = False
        self._flag_italico = False
        self._flag_expandido = False
        self._flag_condensado= False


    @property
    def _colunas(self):
        if self._flag_condensado and not self._flag_expandido:
            return conf.colunas.condensado
        elif self._flag_expandido and not self._flag_condensado:
            return conf.colunas.expandido
        return conf.colunas.normal


    def imprimir(self):
        self.cabecalho()
        self.corpo()
        self.rodape()
        self.fim_documento()


    def centro(self):
        self.impressora.justify_center()
        return self


    def esquerda(self):
        self.impressora.justify_left()
        return self


    def normal(self):
        if self._flag_negrito:
            self.negrito()
        if self._flag_italico:
            self.italico()
        if self._flag_expandido:
            self.expandido()
        if self._flag_condensado:
            self.condensado()


    def negrito(self):
        self._flag_negrito = not self._flag_negrito
        self.impressora.set_emphasized(self._flag_negrito)
        return self


    def italico(self):
        warnings.warn('Estilo "italico" ainda nao disponivel via PyESCPOS',
                stacklevel=2)
        self._flag_italico = not self._flag_italico
        return self


    def expandido(self):
        self._flag_expandido = not self._flag_expandido
        self.impressora.set_expanded(self._flag_expandido)
        return self


    def condensado(self):
        self._flag_condensado = not self._flag_condensado
        self.impressora.set_condensed(self._flag_condensado)
        return self


    def bordas(self, texto_esquerda, texto_direita,
            colunas=None,
            espacamento_minimo=4):
        largura = colunas or self._colunas
        texto = _bordas(texto_esquerda, texto_direita,
                largura=largura,
                espacamento_minimo=espacamento_minimo)
        self.texto(texto)
        return self


    def quebrar(self, texto, colunas=None):
        largura = colunas or self._colunas
        # considera hard-breaks
        linhas_fixas = texto.replace('\r', '').split('\n')
        for linha_fixa in linhas_fixas:
            linhas = textwrap.wrap(linha_fixa, largura)
            for linha in linhas:
                self.texto(linha)
        return self


    def avanco(self, linhas=1):
        self.impressora.lf(lines=linhas)
        return self


    def texto(self, texto):
        self.impressora.text(unidecode(util.forcar_unicode(texto)))
        return self


    def separador(self, caracter='-', colunas=None):
        largura = colunas or self._colunas
        self.texto('-' * largura)
        return self


    def em_condicao_de_teste(self):
        """
        Identifica se o CF-e de venda/cancelamento está em "condição de teste".

        O Manual de Orientação não é claro quanto o que significa "estar em
        condição de teste". Este método irá assumir "condição de teste" quando a
        assinatura (B11) possuir a assinatura de teste, indicada pela constante
        :attr:`satcomum.constantes.ASSINATURA_AC_TESTE`.
        """
        return self.xml.text('infCFe/ide/signAC') == \
                constantes.ASSINATURA_AC_TESTE


    def indicacao_de_teste(self):
        """
        Imprime a indicação de teste se o CF-e estiver em "condição de teste".
        Caso contrário, não faz nada. A indicação de teste, conforme o Manual de
        Orientação deverá ser impressa em itálico (note a linha em branco acima
        e abaixo da inscrição "TESTE"):

        .. sourcecode:: text

                    10        20        30        40      48 :
            ....:....|....:....|....:....|....:....|....:..! :
                                                             :
                             = T E S T E =                   :
                                                             :
            <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< :
            <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< :
            <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< :

        .. note::

            O correto são linhas com sinais de ``>``, mas isso faz com que o
            framework de teste interprete como um bloco de teste (doctest). Os
            sinais de ``<`` dão uma boa ideia do resultado final.

        """
        if self.em_condicao_de_teste():
            self.italico()
            self.avanco()
            self.texto('= T E S T E =')
            self.avanco()
            for i in xrange(3):
                self.texto('>' * conf.colunas.normal)
            self.italico()


    def numero_extrato(self):
        """
        Obtém o número do extrato, elemento ``nCFe`` (B06). Se o CF-e estiver
        em condição de teste retornará ``000000``.

        Veja :meth:`em_condicao_de_teste` para outros detalhes.
        """
        if self.em_condicao_de_teste():
            return '0' * 6
        return self.xml.text('infCFe/ide/nCFe')


    def chave_cfe_code128(self, chave):
        """
        Imprime o código de barras Code128 da chave do CF-e informada.
        """
        code128_params = dict(
                barcode_height=conf.code128_altura,
                barcode_width=escpos.barcode.BARCODE_NORMAL_WIDTH,
                barcode_hri=escpos.barcode.BARCODE_HRI_NONE)

        if conf.code128_quebrar:
            # imprime o Code128 da chave do CF-e em duas partes de 22 digitos
            chave_quebrada = util.partes_chave_cfe(chave, partes=2)
            self.impressora.code128(chave_quebrada[0], **code128_params)
            self.avanco()
            self.impressora.code128(chave_quebrada[1], **code128_params)
        else:
            self.impressora.code128(chave, **code128_params)


    def fim_documento(self):
        """
        Encerra o documento, imprimindo o rodapé (se houver) e avançando ou
        guilhotinando o documento, conforme as configurações.
        """
        self.normal()
        self.avanco()
        self.separador()

        if conf.nota_rodape.esquerda or conf.nota_rodape.direita:
            self.condensado()
            self.bordas(conf.nota_rodape.esquerda, conf.nota_rodape.direita)
            self.condensado()

        if conf.cortar_documento and \
                self.impressora.hardware_features.get(
                        escpos.feature.CUTTER, False):
            self.impressora.cut(partial=conf.cortar_parcialmente)

        elif conf.avancar_linhas > 0:
            self.avanco(conf.avancar_linhas)


    def cabecalho(self):

        self.normal()

        nome_fantasia = self.xml.text('infCFe/emit/xFant', default='')
        razao_social = self.xml.text('infCFe/emit/xNome')

        logradouro = self.xml.text('infCFe/emit/enderEmit/xLgr')
        numero = self.xml.text('infCFe/emit/enderEmit/nro', default='')
        complemento = self.xml.text('infCFe/emit/enderEmit/xCpl', default='')
        bairro = self.xml.text('infCFe/emit/enderEmit/xBairro')

        if numero: # número não é obrigatório
            logradouro = u'{}, {}'.format(logradouro, numero)

        if complemento: # complemento não é obrigatório
            # faz um esforço para manter o complemento na mesma linha que o
            # logradouro/número se couber, senão o complemento irá usar uma
            # linha exclusiva...
            if len(logradouro) + len(complemento) < self._colunas:
                # ok, mantém o complemento na mesma linha que o logradouro
                logradouro = u'{}, {}'.format(logradouro, complemento)
                complemento = '' # ignora a linha que deveria conter o xCpl

        cidade = u'{}/{} CEP: {}'.format(
                self.xml.text('infCFe/emit/enderEmit/xMun'),
                br.uf_pelo_codigo(int(self.xml.text('infCFe/ide/cUF'))),
                br.as_cep(self.xml.text('infCFe/emit/enderEmit/CEP')))

        partes_endereco = [
                logradouro,
                complemento,
                bairro,
                cidade,]

        endereco = u'\r\n'.join([e for e in partes_endereco if e])

        im = 'IM: {}'.format(self.xml.text('infCFe/emit/IM', default=''))
        ie = 'IE: {}'.format(self.xml.text('infCFe/emit/IE'))
        cnpj = 'CNPJ: {}'.format(
                br.as_cnpj(self.xml.text('infCFe/emit/CNPJ')))

        self.centro()
        self.negrito()

        if nome_fantasia:
            self.quebrar(nome_fantasia)

        self.quebrar(razao_social)
        self.negrito()

        self.quebrar(endereco)

        self.avanco()
        self.esquerda()
        self.texto(cnpj)
        self.texto(ie)

        if im:
            self.texto(im)

        self.separador()


    def rodape(self):
        raise NotImplementedError()


    def corpo(self):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
#
# satextrato/cancelamento.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from datetime import datetime

from satcomum import br
from satcomum import util

from .config import conf
from .base import ExtratoCFe


class ExtratoCFeCancelamento(ExtratoCFe):


    def __init__(self, xmlstr, xmlstr_cfe_cancelado, impressora):
        """

        :param xmlstr: String (unicode) contendo o XML do CF-e-SAT de
            cancelamento.

        :param xmlstr_cfe_cancelado: String (unicode) contendo o XML do
            CF-e-SAT de venda para o qual o extrato de cancelamento será
            emitido (o CF-e de venda que foi cancelado).

        :param impressora: Um objeto :class:`escpos.impl.epson.GenericESCPOS`
            ou especialização.
        """
        super(ExtratoCFeCancelamento, self).__init__(xmlstr, impressora)
        self._xmlstr_cfe_cancelado = xmlstr_cfe_cancelado
        self.xml_cfe_cancelado = \
                util.XMLFacadeFromString(self._xmlstr_cfe_cancelado)


    def corpo(self):
        self.corpo_titulo()
        self.corpo_dados_consumidor()
        self.corpo_total_cupom()
        self.corpo_dados_cfe_cancelado()


    def corpo_titulo(self):
        self.normal()
        self.centro()
        self.negrito()
        self.texto('Extrato No. {}'.format(self.numero_extrato()))
        self.texto(u'CUPOM FISCAL ELETRÔNICO - SAT')
        self.texto(u'CANCELAMENTO')
        self.indicacao_de_teste()
        self.negrito()
        self.esquerda()
        self.separador()
        self.negrito()
        self.condensado()
        self.texto(u'DADOS DO CUPOM FISCAL ELETRÔNICO CANCELADO')
        self.condensado()
        self.negrito()


    def corpo_dados_consumidor(self):
        documento = self.xml.text('infCFe/dest/CPF',
                alternative_xpath='infCFe/dest/CNPJ', default='')

        if br.is_cnpjcpf(documento):
            self.normal()
            self.esquerda()
            self.avanco()
            self.texto('CPF/CNPJ do Consumidor: {}'.format(
                    br.as_cnpjcpf(documento)))
        else:
            # apenas normaliza e avança;
            # o total do cupom fiscal cancelado será impresso em seguida;
            self.normal()
            self.avanco()


    def corpo_total_cupom(self):
        self.normal()
        self.esquerda()
        self.negrito()
        self.texto('TOTAL R$ {:n}'.format(
                self.xml.decimal('infCFe/total/vCFe')))
        self.negrito()


    def corpo_dados_cfe_cancelado(self):

        sat_numero_serie = 'SAT no. {}'.format(
                self.xml.text('infCFe/ide/nserieSAT'))

        datahora = datetime.strptime('{}{}'.format(
                self.xml.text('infCFe/dEmi'),
                self.xml.text('infCFe/hEmi')), '%Y%m%d%H%M%S')

        datahora_emissao = datahora.strftime('%d/%m/%Y - %H:%M:%S')

        chave = self.xml.attr('infCFe', 'chCanc')[3:] # ignora prefixo "CFe"
        chave_consulta_partes = ' '.join(util.partes_chave_cfe(chave))

        self.normal()
        self.centro()
        self.avanco()
        self.negrito()
        self.texto(sat_numero_serie)
        self.negrito()
        self.texto(datahora_emissao)

        self.avanco()
        self.esquerda()
        self.condensado()
        self.texto(chave_consulta_partes)
        self.condensado()

        self.avanco()
        self.centro()

        self.chave_cfe_code128(chave)

        self.avanco(2)
        self.impressora.qrcode(util.dados_qrcode(self.xml_cfe_cancelado),
                qrcode_module_size=conf.qrcode.tamanho_modulo,
                qrcode_ecc_level=conf.qrcode.nivel_correcao)



    def rodape(self):

        sat_numero_serie = 'SAT no. {}'.format(
                self.xml.text('infCFe/ide/nserieSAT'))

        datahora = datetime.strptime('{}{}'.format(
                self.xml.text('infCFe/ide/dEmi'),
                self.xml.text('infCFe/ide/hEmi')), '%Y%m%d%H%M%S')

        datahora_emissao = datahora.strftime('%d/%m/%Y - %H:%M:%S')

        chave = self.xml.attr('infCFe', 'Id')[3:] # ignora prefixo "CFe"
        chave_consulta_partes = ' '.join(util.partes_chave_cfe(chave))

        self.normal()
        self.avanco(2)
        self.separador()
        self.centro()
        self.negrito()
        self.quebrar(u'DADOS DO CUPOM FISCAL ELETRÔNICO DE CANCELAMENTO')
        self.avanco()
        self.texto(sat_numero_serie)
        self.negrito()
        self.texto(datahora_emissao)

        self.avanco()
        self.esquerda()
        self.condensado()
        self.texto(chave_consulta_partes)
        self.condensado()

        self.avanco()
        self.centro()

        self.chave_cfe_code128(chave)

        self.avanco(2)
        self.impressora.qrcode(util.dados_qrcode(self.xml),
                qrcode_module_size=conf.qrcode.tamanho_modulo,
                qrcode_ecc_level=conf.qrcode.nivel_correcao)



# -*- coding: utf-8 -*-
#
# satextrato/venda.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import textwrap

from datetime import datetime

from satcomum import br
from satcomum import util

from .config import ZERO
from .config import conf
from .base import _bordas
from .base import ExtratoCFe


class ExtratoCFeVenda(ExtratoCFe):

    anotacoes_antes_obs_contribuinte = None

    anotacoes_corpo = None


    def __init__(self, xmlstr, impressora):
        super(ExtratoCFeVenda, self).__init__(xmlstr, impressora)
        self.anotacoes_antes_obs_contribuinte = []
        self.anotacoes_corpo = []


    def apresentar_item(self, item):
        """
        Apresenta um item (produto/serviço) do CF-e.

        Nesta abordagem, os detalhes do item utilizam uma linha exclusiva,
        abaixo das linhas que mostram o número do item, código e descrição.
        Por exemplo (considerando 48 caracteres por linha):

        .. sourcecode:: text

                    10        20        30        40      48 :
            ....:....|....:....|....:....|....:....|....:..! :
            001 7891234567891 PELLENTESQUE HABITANT MORBI    :
            TRISTIQUE SENECTUS ET NETUS ET MALESUADA FAMES   :
            AC TURPIS EGESTAS VESTIBULUM TORTOR QUAM AMET    :
                                   1 UN x 12,75 (0,30) 12,75 :

        Mesmo no caso em que o nome do produto é muito pequeno, serão utilizadas
        duas linhas para apresentar o item.

        .. sourcecode:: text

                    10        20        30        40      48 :
            ....:....|....:....|....:....|....:....|....:..! :
            001 7891234567891 AB                             :
                                   1 UN x 12,75 (0,30) 12,75 :

        Mas, como é muito rara a situação em que o produto poderia ser
        apresentado em uma única linha, o ganho na facilidade de leitura
        compensa (pelo menos para o consumidor).

        Este método ainda faz um esforço de apresentar o item em apenas uma
        linha tentando "encaixar" o detalhamento do item na mesma linha de
        descrição, desde que caiba. Isso pode ser mais comum, se estiver
        configurado para apresentar os itens em modo condensado (padrão).
        """
        nItem = int(item.root_attr('nItem'))
        cProd = item.text('prod/cProd')
        xProd = item.text('prod/xProd')
        qCom = item.decimal('prod/qCom').normalize()
        uCom = item.text('prod/uCom')
        vUnCom = item.decimal('prod/vUnCom')
        vProd = item.decimal('prod/vProd')
        vItem12741 = item.decimal('imposto/vItem12741', default=ZERO)

        if vItem12741.is_zero():
            detalhe = u'{:n} {:s} x {:n} {:n}'.format(
                    qCom, uCom, vUnCom, vProd)
        else:
            detalhe = u'{:n} {:s} x {:n} ({:n}) {:n}'.format(
                    qCom, uCom, vUnCom, vItem12741, vProd)

        texto = u'{0:03d} {1:s} {2:s}'.format(nItem, cProd, xProd)

        self.normal()
        self.esquerda()

        if conf.itens_modo_condensado:
            self.condensado() # liga

        largura = self._colunas
        linhas = textwrap.wrap(texto, largura)

        ultima_linha = linhas[-1]

        if (len(ultima_linha) + len(detalhe) + 1) > largura:
            # não cabe, o detalhe será apresentado em uma linha exclusiva
            linhas.append(_bordas('', detalhe,
                    largura=largura, espacamento_minimo=1))
        else:
            # cabe; o detalhe será apresentado junto com a última linha
            linhas[-1] = _bordas(ultima_linha, detalhe,
                    largura=largura, espacamento_minimo=1)

        for linha in linhas:
            self.texto(linha)

        if conf.itens_modo_condensado:
            self.condensado() # desliga

        return self


    def corpo(self):
        self.corpo_I_titulo()
        self.corpo_II_consumidor()
        self.corpo_III_legenda()
        self.corpo_IV_itens()
        self.corpo_V_total_cupom()
        self.corpo_VI_meio_pagamento()
        self.corpo_VII_observacoes_fisco()
        self.corpo_VIII_dados_entrega()

        if self.anotacoes_antes_obs_contribuinte:
            self.normal()
            self.esquerda()
            self.avanco()
            for anotacao in self.anotacoes_antes_obs_contribuinte:
                self.texto(anotacao)

        self.corpo_IX_observacoes_contribuinte()

        if self.anotacoes_corpo:
            self.normal()
            self.esquerda()
            self.avanco()
            for anotacao in self.anotacoes_corpo:
                self.texto(anotacao)

        self.anotacoes_antes_obs_contribuinte[:] = []
        self.anotacoes_corpo[:] = []


    def corpo_I_titulo(self):
        self.normal()
        self.centro()
        self.negrito()
        self.texto('Extrato No. {}'.format(self.numero_extrato()))
        self.texto(u'CUPOM FISCAL ELETRÔNICO - SAT')
        self.indicacao_de_teste()
        self.negrito()
        self.esquerda()
        self.separador()


    def corpo_II_consumidor(self):
        # (!) Com a introdução do Anexo I, "Correlação de Campos do Extrato
        #     do CF-e-SAT", parece que ficou claro que o grupo II não deverá
        #     exibir o nome do consumidor, nem se estiver presente.
        #
        #     Como havia implementado originalmente com o nome, apenas
        #     deixei configurável a apresentação do nome do consumidor.
        #
        nome = self.xml.text('infCFe/dest/xNome', default='')
        documento = self.xml.text('infCFe/dest/CPF',
            alternative_xpath='infCFe/dest/CNPJ', default='')

        if br.is_cnpjcpf(documento):
            self.normal()
            self.esquerda()
            self.texto('CPF/CNPJ do Consumidor: {}'.format(
                    br.as_cnpjcpf(documento)))

            if nome and conf.exibir_nome_consumidor:
                self.quebrar(nome)

            self.separador()


    def corpo_III_legenda(self):
        legendas = ' | '.join([
                '#', 'COD', 'DESC', 'QTD', 'UN',
                'VL UN R$', '(VL TR R$)*', 'VL ITEM R$',])

        self.normal()
        self.condensado()
        self.quebrar(legendas)
        self.condensado()
        self.separador()

        self.anotacoes_antes_obs_contribuinte.append(
                u'*Valor aproximado dos tributos do item.')


    def corpo_IV_itens(self):

        for item in self.xml.iterate('infCFe/det'):
            self.apresentar_item(item)

            vProd = item.decimal('prod/vProd')
            vDesc = item.decimal('prod/vDesc', default=ZERO)
            vRatDesc = item.decimal('prod/vRatDesc', default=ZERO)
            vOutro = item.decimal('prod/vOutro', default=ZERO)
            vRatAcr = item.decimal('prod/vRatAcr', default=ZERO)

            # (!) Do modo como está implementado, descontos e acréscimos
            # serão ambos tratados como se pudessem coexistir, embora o
            # Manual de Orientação seja claro quando diz que "valores de
            # descontos e acréscimos são mutuamente exclusivos".
            forcar_exibir = False

            if not vDesc.is_zero() or forcar_exibir:
                self.bordas('desconto sobre item', '- {:n}'.format(vDesc))

            if not vRatDesc.is_zero() or forcar_exibir:
                self.bordas('rateio de desconto sobre subtotal',
                        '- {:n}'.format(vRatDesc))

            if not vOutro.is_zero() or forcar_exibir:
                self.bordas(u'acréscimo sobre item', '+ {:n}'.format(vOutro))

            if not vRatAcr.is_zero() or forcar_exibir:
                self.bordas(u'rateio de acréscimo sobre subtotal',
                        '+ {:n}'.format(vRatAcr))

            # item tributado pelo ISSQN?
            if item.text('imposto/ISSQN/cNatOp', default='@') != '@':
                # o campo cNatOp (U09) retornou algo diferente do valor default
                # indicado; logo, assume que o item foi tributado pelo ISSQN...
                vBC = item.decimal('imposto/ISSQN/vBC', default=ZERO)
                vDeducISSQN = item.decimal('imposto/ISSQN/vDeducISSQN',
                        default=ZERO)

                if not vDeducISSQN.is_zero():
                    self.bordas(u'dedução para ISSQN',
                            '- {:n}'.format(vDeducISSQN))

                self.bordas(u'base de cálculo ISSQN', '{:n}'.format(vBC))


    def corpo_V_total_cupom(self):

        # calcula o total de descontos apenas dos itens tributados por ISSQN,
        # usando os campos vDesc (I12) e vOutro (I13);
        issqn_total_desc = ZERO
        issqn_total_acres = ZERO

        for item in self.xml.iterate('infCFe/det'):
            # item tributado pelo ISSQN?
            if item.text('imposto/ISSQN/cNatOp', default='@') != '@':
                # o campo cNatOp (U09) retornou algo diferente do valor default
                # indicado; logo, assume que o item foi tributado pelo ISSQN...
                issqn_total_desc += item.decimal('prod/vDesc', default=ZERO)
                issqn_total_acres += item.decimal('prod/vOutro', default=ZERO)

        # total de descontos sobre itens (vDesc, W05);
        # total de outras despesas acessórias sobre itens (vOutro W10);
        total_desc = self.xml.decimal(
                'infCFe/total/ICMSTot/vDesc', default=ZERO) + issqn_total_desc

        total_acres = self.xml.decimal(
                'infCFe/total/ICMSTot/vOutro', default=ZERO) + issqn_total_acres

        # vProd (W04)
        vProd = self.xml.decimal('infCFe/total/ICMSTot/vProd')
        vDescSubtot = self.xml.decimal(
                'infCFe/total/DescAcrEntr/vDescSubtot', default=ZERO)
        vAcresSubtot = self.xml.decimal(
                'infCFe/total/DescAcrEntr/vAcresSubtot', default=ZERO)

        self.normal()
        self.avanco()

        forcar_exibir = False

        ha_desc_itens = not total_desc.is_zero() or forcar_exibir
        ha_desc_subtotal = not vDescSubtot.is_zero() or forcar_exibir
        ha_desconto = ha_desc_itens or ha_desc_subtotal or forcar_exibir

        ha_acres_itens = not total_acres.is_zero() or forcar_exibir
        ha_acres_subtotal = not vAcresSubtot.is_zero() or forcar_exibir
        ha_acrescimo = ha_acres_itens or ha_acres_subtotal or forcar_exibir

        if ha_desconto or ha_acrescimo:
            # há desconto ou acréscimo sobre os itens e/ou sobre o subtotal
            self.bordas('Total bruto de itens', '{:n}'.format(vProd))

        if ha_desc_itens:
            # há desconto sobre itens
            self.bordas('Total de descontos sobre item',
                    '- {:n}'.format(total_desc))

        if ha_desc_subtotal:
            # há desconto sobre o subtotal
            self.bordas('Desconto sobre subtotal', '- {:n}'.format(vDescSubtot))

        if ha_acres_itens:
            # há acréscimo sobre itens
            self.bordas(u'Total de acréscimos sobre item',
                    '- {:n}'.format(total_acres))

        if ha_acres_subtotal:
            # há acréscimo sobre o subtotal
            self.bordas(u'Acréscimo sobre subtotal',
                    '- {:n}'.format(vAcresSubtot))

        vCFe = self.xml.decimal('infCFe/total/vCFe')

        self.negrito()
        self.bordas('TOTAL R$', '{:n}'.format(vCFe))
        self.negrito()


    def corpo_VI_meio_pagamento(self):
        self.normal()
        self.avanco()

        for pagto in self.xml.iterate('infCFe/pgto/MP'):
            self.bordas(util.meio_pagamento(pagto.text('cMP')),
                    '{:n}'.format(pagto.decimal('vMP')))

        valor_troco = self.xml.decimal('infCFe/pgto/vTroco')
        if not valor_troco.is_zero():
            self.bordas('Troco R$', '{:n}'.format(valor_troco))


    def corpo_VII_observacoes_fisco(self):
        """
        Grupo VII, Observações do Fisco. Redação atual, efeitos até 31-12-2016.
        Usa o atributo ``xCampo`` (Z04) e o elemento ``xTexto`` (Z05):

        .. sourcecode:: xml

            <CFe>
              <infCFe>
                <infAdic>
                  <obsFisco xCampo="Foo">
                      <xTexto>Conteúdo de Foo</xTexto>
                  </obsFisco>
                  <obsFisco xCampo="Bar">
                      <xTexto>Conteúdo de Bar</xTexto>
                  </obsFisco>
                </infAdic>
              </infCFe>
            </CFe>

        Nova redação, efeitos a partir de 01-01-2017.
        Usa o atributo ``xCampo`` (ZA02) e o elemento ``xTexto`` (ZA03):

        .. sourcecode:: xml

            <CFe>
              <infCFe>
                <obsFisco xCampo="Spam">
                    <xTexto>Conteúdo de Spam</xTexto>
                </obsFisco>
                <obsFisco xCampo="Eggs">
                    <xTexto>Conteúdo de Eggs</xTexto>
                </obsFisco>
              </infCFe>
            </CFe>

        """
        iniciado = False

        for obs in self.xml.iterate('infCFe/infAdic/obsFisco'):

            if not iniciado:
                self.normal()
                self.esquerda()
                self.avanco()
                self.condensado()
                iniciado = True

            self.quebrar(u'{}: {}'.format(
                    obs.root_attr('xCampo'),
                    obs.text('xTexto')))

        if iniciado:
            self.condensado() # desliga


    def corpo_VIII_dados_entrega(self):

        logradouro = self.xml.text('infCFe/entrega/xLgr', default='')
        if not logradouro:
            return

        # há endereço de entrega;
        # prepara para imprimir, por que será necessário considerar o número
        # de colunas para construir o endereço o mais legível possível;
        self.normal()

        # uma vez que há o logradouro, então existe um endereço de entrega;
        # se existe um endereço de entrega então os únicos elementos que
        # NÃO SÃO OBRIGATÓRIOS são `nro` (G03) e `xCpl` (G04);
        numero = self.xml.text('infCFe/entrega/nro', default='')
        complemento = self.xml.text('infCFe/entrega/xCpl', default='')
        bairro = self.xml.text('infCFe/entrega/xBairro')

        if numero: # número não é obrigatório
            # mas existe, então o coloca próximo ao logradouro
            logradouro = u'{}, {}'.format(logradouro, numero)

        if complemento: # complemento não é obrigatório
            # faz um esforço para manter o complemento na mesma linha que o
            # logradouro/número se couber, senão o complemento irá usar uma
            # linha exclusiva...
            if len(logradouro) + len(complemento) < self._colunas:
                # ok, mantém o complemento na mesma linha que o logradouro
                logradouro = u'{}, {}'.format(logradouro, complemento)
                complemento = '' # ignora a linha que deveria conter o xCpl

        cidade = u'{}/{}'.format(
                self.xml.text('infCFe/entrega/xMun'),
                self.xml.text('infCFe/entrega/UF'))

        partes_endereco = [logradouro, complemento, bairro, cidade,]
        endereco = '\r\n'.join([e for e in partes_endereco if e])

        self.esquerda()
        self.separador()
        self.negrito()
        self.texto('DADOS PARA ENTREGA')
        self.negrito()
        self.quebrar(u'Endereço: {}'.format(endereco))

        nome_destinatario = self.xml.text('infCFe/dest/xNome', default='')
        if nome_destinatario:
            self.quebrar(u'Destinatário: {}'.format(nome_destinatario))


    def corpo_IX_observacoes_contribuinte(self):
        infCpl = self.xml.text('infCFe/infAdic/infCpl', default='')
        vCFeLei12741 = self.xml.decimal(
                'infCFe/total/vCFeLei12741', default=ZERO)

        if infCpl or not vCFeLei12741.is_zero():
            self.normal()
            self.esquerda()
            self.separador()
            self.negrito()
            self.texto(u'OBSERVAÇÕES DO CONTRIBUINTE')
            self.negrito()

            if infCpl:
                self.condensado()
                self.quebrar(infCpl)
                self.condensado()

            if not vCFeLei12741.is_zero():
                self.condensado()
                self.quebrar(u'Valor aproximado dos tributos deste cupom')
                self.bordas(
                        u'(Lei Fed. 12.741/2012) R$',
                        '{:n}'.format(vCFeLei12741))
                self.condensado()


    def rodape(self):

        sat_numero_serie = 'SAT no. {}'.format(
                self.xml.text('infCFe/ide/nserieSAT'))

        datahora = datetime.strptime('{}{}'.format(
                self.xml.text('infCFe/ide/dEmi'),
                self.xml.text('infCFe/ide/hEmi')), '%Y%m%d%H%M%S')

        datahora_emissao = datahora.strftime('%d/%m/%Y - %H:%M:%S')

        chave_cfe = self.xml.attr('infCFe', 'Id')[3:] # ignora prefixo "CFe"
        chave_consulta_partes = ' '.join(util.partes_chave_cfe(chave_cfe))

        self.normal()
        self.separador()
        self.centro()
        self.negrito()
        self.texto(sat_numero_serie)
        self.negrito()
        self.texto(datahora_emissao)

        self.avanco()
        self.esquerda()
        self.condensado()
        self.texto(chave_consulta_partes)
        self.condensado()

        self.avanco()
        self.centro()

        self.chave_cfe_code128(chave_cfe)

        self.avanco(2)
        self.impressora.qrcode(util.dados_qrcode(self.xml),
                qrcode_module_size=conf.qrcode.tamanho_modulo,
                qrcode_ecc_level=conf.qrcode.nivel_correcao)



# -*- coding: utf-8 -*-
#
# satextrato/config.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from decimal import Decimal

import escpos.barcode

ZERO = Decimal(0)


class LarguraBobina(object):
    # FIXME: A implementação de ESC/POS deverá saber quais os valores para os
    #        modos normal, condensado e expandido;
    #
    normal = 48
    """
    Número de caracteres por linha no modo de impressão normal.
    """

    condensado = 57
    """
    Número de caracteres por linha no modo de impressão condensado.
    """

    expandido = 24
    """
    Número de caracteres por linha no modo impressão expandido.
    """


class ParametrosQRCode(object):
    tamanho_modulo = escpos.barcode.QRCODE_MODULE_SIZE_4
    """
    Determina o tamanho do módulo QRCode.
    Valor padrão é :attr:`escpos.barcode.QRCODE_MODULE_SIZE_4`
    """

    nivel_correcao = escpos.barcode.QRCODE_ERROR_CORRECTION_L
    """
    Determina o nível de correção (ECC) do QRCode.
    Valor padrão é :attr:`escpos.barcode.QRCODE_ERROR_CORRECTION_L`
    """


class NotaRodape(object):

    esquerda = 'Extrato SAT-CF-e'
    """
    Texto alinha à borda esquerda da bobina.
    """

    direita = 'http://git.io/vJRRk'
    """
    Texto alinhado à borda direita da bobina.

    .. todo::

        Usar https://github.com/blog/985-git-io-github-url-shortener
    """


class ConfiguracoesExtrato(object):

    colunas = LarguraBobina()
    """
    Referência aos atributos que determinam a largura da bobina,
    classe :class:`LarguraBobina` ou implementação especializada.
    """

    qrcode = ParametrosQRCode()
    """
    Referência aos atributos que determinam os parâmetros do QRCode,
    classe :class:`ParametrosQRCode` ou implementação especializada.
    """

    nota_rodape = NotaRodape()
    """
    Referência aos atributos que serão usados na impressão da nota de rodapé,
    classe :class:`NotaRodape` ou implementação especializada.
    """

    itens_modo_condensado = True
    """
    Indica se os dados dos itens, no corpo do Extrato do CF-e-SAT deverão ser
    impressos em modo condensado. Se não, serão impressos em modo normal.
    """

    avancar_linhas = 10
    """
    Número de linhas em branco a avançar ao final do Extrato do CF-e-SAT.

    Esta configuração será considerada apenas se :attr:`cortar_documento`
    estiver configurado como ``False``. Para não cortar nem avançar, deixe
    :attr:`cortar_documento` como ``False`` e :attr:`avancar_linhas` em ``0``.
    """

    cortar_documento = True
    """
    Se o documento deve ser cortado ao ser concluído. Esta configuração não
    terá efeito se o equipamento não possuir uma guilhotina. Se o equipamento
    não possuir uma guilhotina, então :attr:`avancar_linhas` será respeitado.
    """

    cortar_parcialmente = True
    """
    Ao cortar o documento (:attr:`cortar_documento`), esta propriedade indicará
    se o corte deverá ser parcial ou total.
    """

    exibir_nome_consumidor = False
    """
    Indica se o nome do consumidor deve ou não ser exibido no Grupo II
    "Dados do Consumidor" do corpo do Extrato do CF-e-SAT. Esta configuração
    será considerada apenas se o elemento ``xNome`` (E04) existir no CF-e.

    .. warning::

        A documentação para impressão do extrato não diz que o nome do
        consumidor deverá ser exibido, mas que apenas o documento (CNPJ/CPF)
        deverão ser exibidos.

    """

    code128_quebrar = False
    """
    Indica se o código de barras Code128 que representa a chave do CF-e deverá
    ser quebrado em 2 partes de 22 dígitos cada ou não. Essa decisão depende
    do equipamento utilizado para impressão.
    """

    code128_altura = 96
    """
    Indica a altura preferencial em milímetros para código de barras que
    representa a chave do CF-e. A altura poderá ser ignorada se o equipamento
    utilizado para impressão não suportar esta configuração.

    A altura padrão é ``96``, que é calculado com base na média de 0,125mm por
    ponto. Assim, ``96`` significa uma altura de aproximadamente 12mm.
    """

    cliche = None
    """
    .. warning::

        Rascunho. Não implementado.

    Dados para o clichê de quaisquer Extratos CF-e-SAT, sejam eles de venda
    ou de cancelamento. Os dados do clichê serão usados como uma forma de
    melhorar a formatação das informações. Essas informações, pelo menos em
    tese, são as mesmas informações que se obteria para produzir o clichê a
    partir do CF-e.

    Trata-se de uma lista de strings. Cada string será uma linha do clichê que
    poderá indicar certos aspectos da formatação através de *hashtags*, como
    por exemplo, ``#negrito``, ``#condensado`` e ``#expandido``.

    Hashtags conflitantes serão ignoradas e vencerá a última hashtag carregada.
    Por exemplo, se houver a especificação ``#condensado`` e, sem seguida,
    a hashtag ``#expandido``, será considerado expandido se esta for a última
    hashtag lida.

    Um exemplo de especificação do clichê::

        conf.cliche = [
                u'#expandido #negrito BASE4 SISTEMAS',
                u'#negrito Base4 Sistemas Ltda ME',
                u'#condensado Rua Armando Gulim, 65',
                u'#condensado Parque Glória III',
                u'#condensado Catanduva/SP - CEP 15807-250',]

    """


conf = ConfiguracoesExtrato()
"""
Variável de módulo utilizada como base para as configurações de impressão.
"""
